function version = common_version()
  version = "1.5.4";

  // 1.5.3 -> 1.5.4
  // - Created common.sci.
  // - Moved checksize, {row,col}Map from me.sci.
  // - Added progress bar code (was originally in demos/demo.sci).
  // - Added custom error code (cerror).
  // - Added an isEqual function to compare equality of multiple matrices easily.
  // - Added a sumFinites to sum the finite numbers in a (hyper)matrix.
endfunction

function [b,s] = checksize(n,m,sizef,functionname)
  //assume n,m positive integer, and test whether sizef=[n,m]
  //internal function

  b = n==sizef(1) & m==sizef(2);s='';
  if ~b
    if n~=sizef(1)
      s=sprintf("Unconsistent size of f in %s. size(Xr)=%i, but size(f,1)=%i",functionname,n,sizef(1));
    end;
    if m~=sizef(2)
      s=sprintf("Unconsistent size of f in %s. size(Xc)=%i, but size(f,2)=%i",functionname,m,sizef(2));
    end;
  end;
endfunction

function b = isEqual(varargin)
  // Compares whether a variable number of matrices are equal, by
  // comparing successive pairs of matrices via isEqual2.  See
  // help isEqual2 for more information.
  n = length(varargin);
  b = %t;
  for i = 1:n-1
    if (~isEqual2(varargin(i), varargin(i+1))) then
      b = %f;
      return;
    end;
  end;
endfunction

function b = isEqual2(x, y)
  // Compares whether two matrices are equal, including infinities,
  // and, unlike isequal, tests floating point numbers properly.
  // Returns a scalar.  Returns false if the matrix contains NaNs.

  if (or(sign(x) ~= sign(y))) then
    b = %f;
  else
    b = and(abs(x - y) < 1E-6 | (isinf(x) & isinf(y)));
  end;
endfunction

function n = sumFinites(x)
  // Returns the sum of the finite numbers in the matrix x.

  b = ~isinf(x) & ~isnan(x);
  n = sum(x(b));
endfunction

function out=rowMap(func,m)
  // Map a function to each of the rows of a matrix, putting the resulting row vectors into the output matrix
  // @param func A function that takes two parameters (the first is a row vector,
  //        the second the number of the current row) and returns a row vector

  out=[]
  s=size(m)
  n=s(1)
  for i=1:n
    out(i,:)=func(m(i,:),i)
  end
endfunction

// Map a function to each of the columns of a matrix, putting the resulting column vectors into the output matrix
// @param func A function that takes two parameters (the first is a column vector,
//        the second the number of the current column) and returns a column vector
function out=colMap(func,m)
  out=[]
  s=size(m)
  n=s(2)
  for i=1:n
    out(:,i)=func(m(:,i),i)
  end
endfunction

// All CCA functions should use this function instead of error(),
// so that we can act differently depending on when we're generating
// an error to pass a test or generating one for real.  Currently
// this function could be eliminated by having __testing_errors = 9999
// in test.sci:errorTestWrapper(), but this way presents more flexibility.
function cerror(msg)
  global __testing_errors;
  if (__testing_errors) then
    error(msg, __testing_errors);
  else
    error(msg);
  end;
endfunction

function beginProgress(width)
  global defaultProgressWidth;
  if (defaultProgressWidth == []) then defaultProgressWidth = 20; end;

  if argn(2) < 1 then
    showProgress(0);
  else
    showProgress(0, width);
  end;
endfunction

function showProgress(percent, width, altstr)
  // [...] 100%
  // 1 2 345678
  global defaultProgressWidth;
  if argn(2) < 2 then width = defaultProgressWidth;
  elseif width == %f then width = defaultProgressWidth;
  end;
  percent = min(percent, 1);
  width = max(width, 8);
  barWidth = width - 7;
  starCnt = int(percent * barWidth);
  spaceCnt = barWidth - starCnt;
  printf("\r[");                     // '\r' is also code2str(113)
  for i = 1:starCnt; printf("*"); end;
  for i = 1:spaceCnt; printf("_"); end;
  printf("] %3d%%", int(percent*100));
  if argn(2) >= 3 then printf("  %s", altstr); end;
endfunction

function finishProgress(width, altstr)
  if argn(2) == 0 then
    showProgress(1);
  elseif argn(2) == 1 then
    showProgress(1, width);
  else
    showProgress(1, width, altstr);
  end;
  printf("\n");
endfunction

function abortProgress()
  printf(" -Cancelled-\n");
endfunction
