// op.sci loads functions into the CCA toolbox to run computations on
// operators.  Currently there are algorithms to compute Fitzpatrick
// functions of given, finite or infinite, orders.  All of the main
// functions in this file begin with op_.

function version = op_version()
  version = "1.5.4";

  // 1.5.3 -> 1.5.4
  // - Created op.sci.
  // - Implemented four Fitzpatrick algorithms:
  //      op_fitz_brute
  //      op_fitz_direct
  //      op_fitz
  //      op_fitzinf
  // - Added unit tests to reach 100% coverage of op.sci.
endfunction

function F = op_fitzinf(B, x, xstar)
  // Evaluate the Fitzpatrick function F[A,%inf](x, xstar) via
  // Rockafellar functions using an O(m^3) algorithm when m==Nx==Nx*.
  // The code to compute one PLQ function is the same as that in
  // plq.sci:plq_rock().
  //
  // F[A, %inf](x, xstar) = sup [a in Dom A] (a*xstar + R[A,a](x))
  // where B: x -> conv(A*x): union(i=1:m, {a(i)} cross [bminus(i), bplus(i)].
  //
  // B = [a; bminus; bplus];    // a, bminus, bplus are row vectors.

  if (size(x, 1) < size(x, 2)) then x = x'; end;
  if (size(xstar, 1) < size(xstar, 2)) then xstar = xstar'; end;

  if (~and(B(1,2:$) - B(1,1:$-1) >= 0)) then
    [a,k] = gsort(B(1,:), "c", "i");
    B = [a; B(2:$,k)];
  end;

  if (or(B(1,2:$) - B(1,1:$-1) == 0)) then
    cerror("op_fitzinf: a values are not strictly increasing.");
    F = [];
    return;
  end;

  // Is the operator not monotonically nondecreasing?
  if (~(and(B(3,:)-B(2,:) >= 0) & and(B(2,2:$)-B(3,1:$-1) >= 0))) then
    F = %inf * ones(size(x, 1), size(xstar, 1));
    return;
  end;

  //          1         2:m+1        m+2
  m = size(B, 2);
  a = [-%inf, B(1,:), %inf];
  bm = [%nan, B(2,:), %nan];
  bp = [%nan, B(3,:), %nan];
  R = [];

  // Compute k=1:m functions R[A,a_k](x), each having i=1:m terms.
  // O(m(m + Nx)) = O(m^2 + Nx)
  for k = 2:m+1                                   // O(m)
    plqf = [];
    xi = 2;
    // Precompute a sum to improve the efficiency of repeatedly
    // computing Rockafellar functions.
    Sminus = sum((a(2:k-1) - a(3:k)) .* bm(3:k)); // ..O(m)
    Splus = 0;

    for i = 2:m+1                                 // ..O(m)
      if (i <= k) then
        b = bm(i);
        c = -a(i)*bm(i) + Sminus;
        Sminus = Sminus - (a(i) - a(i+1))*bm(i+1);
        plqf = [plqf; a(xi), 0, b, c];
        xi = xi + 1;
      end;

      if (i >= k) then
        b = bp(i);
        c = -a(i)*bp(i) + Splus;
        Splus = Splus + (a(i+1) - a(i))*bp(i);
        plqf = [plqf; a(xi), 0, b, c];
        xi = xi + 1;
      end;
    end;

    //printf("op_fitzinf, plq function #%d:", k);
    //disp(plqf);
    R = [R; -plq_eval(plqf, x)'];                 // ..O(m + Nx)
  end;

  //disp(-R);
  F = zeros(size(x,1), size(xstar,1));
  a = B(1,:)';
  Nx = size(x, 1);
  for i = 1:Nx                           // O(Nx)
    F(i,:) = lft_llt(a, R(:,i), xstar)'; // ..O(m + Nxstar)
  end;
endfunction

function F = op_fitz(A, n, x, xstar)
  // Evaluate the Fitzpatrick function F[A,n](x, xstar) using
  // a recursive algorithm that uses LLT to achieve quartic
  // complexity O(m^3) when m == Nx = Nx*.
  //
  // A = [a; astar];

  if (size(x,1) < size(x,2)) then x = x'; end;
  if (size(xstar,1) < size(xstar,2)) then xstar = xstar'; end;

  if (~and(A(1,2:$) - A(1,1:$-1) >= 0)) then
    [a,k] = gsort(A(1,:), "c", "i");
    A = [a; A(2,k)];
  end;

  Nx = size(x, 1);
  Nxstar = size(xstar, 1);
  m = size(A, 2);

  a = A(1,:)';
  astar = A(2,:)';

  if (n == %inf) then
    // Is the operator monotonically nondecreasing?
    if (and(astar(2:$) - astar(1:$-1) >= 0)) then
      n = m + 1;
    else
      F = %inf * ones(size(x, 1), size(xstar, 1));
      return;
    end;
  end;

  select (n)

  case 1 then
    F = x * xstar';

  case 2 then
    F = %inf * ones(m, m);
    for i = 1:m
      F(i,i) = a(i)*astar(i);
    end;
    F = lft_llt2d(astar, a, F, x, xstar);

  case 3 then
    f0 = -%inf * ones(m, m);
    for i = 1:m
      for j = 1:m
        //v = -a(i)*astar(i) - a(j)*astar(j) + astar(i)*a(j);
        v = a(i)*astar(i) + a(j)*astar(j) - astar(i)*a(j);
        f0(i,j) = max(f0(i,j), v);
      end;
    end;

    F = lft_llt2d(astar, a, f0', x, xstar);
    //F = lft_llt2d(a, astar, f0, x, xstar)';

  else
    F2 = op_fitz(A, n-2, a, astar); // F_{A,n-2}, m x m

    f = zeros(m, m);
    for i = 1:m          // a_1 = a(i)
      a1 = a(i);
      for j = 1:m        // a_{n-1}^* = astar(j)
        anstar = a(j);
        //f(i,j) = -a(i)*astar(i) - a(j)*astar(j) + a(j)*astar(i);
        f(i,j) = -a(i)*astar(i) - a(j)*astar(j) + F2(j, i);
      end;
    end;

    F = lft_llt2d(astar, a, -f', x, xstar);

  end;
endfunction

function F = op_fitz_direct(A, n, x, xstar)
  // Evaluate the Fitzpatrick function F[A,n](x, xstar) using
  // a recursive algorithm with quartic complexity O(m^4).
  //
  // A = [a; astar];

  if (size(x,1) < size(x,2)) then x = x'; end;
  if (size(xstar,1) < size(xstar,2)) then xstar = xstar'; end;

  if (~and(A(1,2:$) - A(1,1:$-1) >= 0)) then
    [a,k] = gsort(A(1,:), "c", "i");
    A = [a; A(2,k)];
  end;

  if (n == %inf) then
    // Is the operator monotonically nondecreasing?
    if (and(A(2,2:$) - A(2,1:$-1) >= 0)) then
      n = size(A, 2) + 1;
    else
      F = %inf * ones(size(x, 1), size(xstar, 1));
      return;
    end;
  end;

  F = op_fitz_direct_do(A, n, x, xstar);
endfunction

function F = op_fitz_direct_do(A, n, x, xstar)
  if (n == 1) then
    F = x * xstar';
    return;
  end;

  Fprev = op_fitz_direct_do(A, n - 1, A(1,:)', xstar);
  m = size(A, 2);
  Nx = size(x, 1);
  Nxstar = size(xstar, 1);
  F = zeros(Nx, Nxstar);

  for i = 1:Nx
    xi = x(i);
    for j = 1:Nxstar
      xstarj = xstar(j);

      sup = -%inf;
      for k = 1:m
        z = Fprev(k,j) + xi*A(2,k) - A(1,k)*A(2,k);
        sup = max(sup, z);
      end;
      F(i, j) = sup;
    end;
  end;
endfunction

function F = op_fitz_brute(A, n, x, xstar)
  // Evaluate the Fitzpatrick function F[A,n](x, xstar) using
  // a brute-force method with complexity O(n*m^(n-1)*N).
  //
  // A = [a; astar];

  if (size(x, 1) < size(x, 2)) then x = x'; end;
  if (size(xstar, 1) < size(xstar, 2)) then xstar = xstar'; end;

  Nx = size(x, 1);
  Nxstar = size(xstar, 1);
  m = size(A, 2);

  if (~and(A(1,2:$) - A(1,1:$-1) >= 0)) then
    [a,k] = gsort(A(1,:), "c", "i");
    A = [a; A(2,k)];
  end;

  if (n == %inf) then
    // Is the operator monotonically nondecreasing?
    if (and(A(2,2:$) - A(2,1:$-1) >= 0)) then
      n = m + 1;
    else
      F = %inf * ones(size(x, 1), size(xstar, 1));
      return;
    end;
  end;

  if (n == 1) then
    F = x * xstar';
    return;
  end;

  F = zeros(Nx, Nxstar);
  a1 = ones(1, n-1);

  // Todo: checksize()
  for i = 1:Nx
    xi = x(i);
    for j = 1:Nxstar
      xstarj = xstar(j);

      a = [A(1,1)*a1; A(2,1)*a1; a1];     // Each column = [A(:, i); i]
      sup = -%inf;
      permuteA = %t;
      while (permuteA)
        z = (a(1,2:(n-1)) - a(1,1:(n-2))) .* a(2,1:(n-2));
        z = sum(z) + (xi - a(1,n-1)) * a(2,n-1) + a(1,1) * xstarj;
        sup = max(sup, z);

        // Permute A with a loop to avoid recursion.
        k = 0;
        wrap = %t;
        while (wrap & k < n)
          k = k + 1;
          if (k == n) then break; end;
          z = a(3,k) + 1;
          if (z <= size(A,2)) then
            wrap = %f;
          else
            wrap = %t;
            z = 1;
          end;
          a(1,k) = A(1,z);
          a(2,k) = A(2,z);
          a(3,k) = z;
        end;

        if (k == n) then permuteA = %f; end;
      end;

      F(i,j) = sup;
    end;
  end;
endfunction
