//////// Coverage testing library for Scilab
//
// Author: Bryan Gardiner

// This library makes use of Scilab's profile() function to perform
// coverage testing almost transparently over-top of your existing
// unit testing code, while testing for error conditions as well.

// Currently, function statement coverage is provided.  Because
// Scilab's profiling works by lines, not statements, we get statement
// coverage by reformatting source code to contain one statement per
// line.  This reformatted version (filename + "_coverage") is kept
// cached and only rebuilt when necessary.

// Reporting and increasing coverage is also made easy.  Functions
// return line numbers that refer back to absolute line numbers in
// your source files, instead of Scilab's function-relative numbering.

// Unfortunately, because Scilab's syntax is somewhat free-form, with
// lots of quirks, getting accurate statement coverage is difficult.
// (A few examples are ' working as transpose and as a string delimiter,
// "then" being required only in some cases, and statements which Scilab
// will never mark as executed.)

// As well, currently (Scilab 4.1.2) there is no way to load a
// file for profiling without using getf(..., "p"), which doesn't
// recognize newer syntax such as top-level statements or nested
// functions.  Scilab 5 will have add_profiling(), reset_profiling(),
// remove_profiling(), and recompilefunction().

//////// Adding coverage testing to your unit tests:

// . In your main test script, before the tests start running, add code to:
//   . Load coverage.sci.  Both exec and getf work.
//   . Call cover_reset().  This sets the coverage library back to
//     a state where no profiled files or functions have been loaded,
//     to its knowledge.
//   . Call cover_profileFiles() to make sure your _coverage files are up to date.
//
// . Load your files, one at a time:
//     clear cover_loadedFile;                     // Prevent redefinition warnings.
//     getf(cover_profiledName("file1.sci"), "p"); // Load the profiled code.
//     cover_newFile();                     // Tell coverage.sci about the new file.
//
// . Run your tests as usual!

//////// Function reference:
//
// **NOTE**   All functions have their names prefixed with   **NOTE**
// **NOTE**      "cover_", which is omitted for brevity.     **NOTE**
//
// See the functions themselves for in-depth documentation.

// The following functions are used for reporting coverage:
//
//     results()
//         Returns a struct containing overall collected data.
//     printStmtCoverage(cov?, unit?, dbg?)
//         Prints a statement coverage report for all loaded functions.
//     printFn(name)
//         Prints a function coverage report, including line ranges not executed.
//     fnStmtCoverage(name)
//         Returns statement coverage information about one function.
//     missedOrigLines = fnLinesMissed(name)
//         Returns a vector of lines not executed in a function.

// The following functions deal with profiling and loading user code:
//
//     reset()
//         Resets the coverage library back to its initial state (nothing loaded).
//     [filesProfiled, success, reason] = profileFiles(filenames?, forceProfiling?)
//         Creates reformatted versions of files for profiling.
//     newFile()
//         Should be called after loading a profiled file.
//     profiledFileSuffix()
//         The suffix the library uses for profiled files.
//     profiledName = profiledFile(filename)
//         Returns the profiled name of a non-profiled file.
//     profiledFiles()
//         Returns a vector of the non-profiled names of profiled files.
//     loadedFiles()
//         Returns a vector of the non-profiled names of loaded, profiled files.
//     isFileProfiled(filename)
//         Returns whether a profiled version of a non-profiled file exists.
//     loadedFns()
//         Returns all the functions loaded for coverage testing.

// The following functions allow unit tests to test for error conditions:
//
//     coverErrors(bool)
//         Return or set whether errors should be caught by the coverage tester.
//     bool = errored()
//         Checks whether or not an error has occurred.
//     error(x, y)
//         Signals an error, which may not be caught by the coverage tester.

// The following are utility functions that most users shouldn't need to call:
//
//     [reformatted, origLines, success, reason] = reformatFile(fsrc)
//         Reformats the lines of fsrc, keeping track of original line numbers.
//     [success, reason] = writeProfiledFile(src, origLines, baseName, fp)
//         Writes a profiled file from the results of reformatFile().
//     name = fnNameFromDecl(line)
//         Finds just the function name in a function declaration.
//     b = isBegOfNameChar(ch)
//         Controls what characters start an identifier.
//     b = isNameChar(ch)
//         Controls what characters can exist in an identifier after the first char.
//     b = isAnyNameChar(ch)
//         Whether the char is a identifier start char or continuation char.
//     b = isTranspose(prevCh)
//         Whether a ' following prevCh is a transpose (or a string delimiter).
//     b = isNonex(stmt)
//         Whether a statement is a nonex.
//     internalError(reason)
//         Displays an error message.
//     loadedFile()
//         Helps initializes a loaded profiled file.  Defined in profiled files.

//////// Technical stuff:

// Little gotchas in Scilab syntax:
//   . Line continuations: two or more periods at the end of a line
//       joins the next line to the current one.  There is no token
//       boundary, i.e. "2...\n3" is interpreted as the number 23.
//       Line continuations do not work in a comment.  They do work
//       in strings.
//   . Distinguishing between ' string and ' transpose is done by
//       looking comparing the previous character to a set of allowed
//       transpose characters.  Space is excluded.  Digits are included.
//   . Funky string syntax:
//       Scilab "A""Z" -> C "A\"Z"    ; Inside character counts.
//              "A''Z" ->   "A'Z"     ; Inside character still counts.
//              "A"'Z" ->   "A'Z"     ; *Right* inside character counts.
//              "A'"Z" ->   "A\"Z"    ; Same.
//              "AZ'   ->   "AZ"      ; Also, Scilab doesn't care which you start
//              'AZ"   ->   "AZ"      ;  or end a string with.
//   . Brackets [] and braces {} behave exactly the same.  They have
//     the same weirdness that string delimiters do.
//   . Two statement delimiters, both of which are overloaded.
//       Not so bad, just have to make sure we're not in parentheses
//       before going on to a new line.
//   . Inconsistent requirement of "then":
//       if test then, foo; end;    // Perfect! Split line on ","
//       if test, foo; end;         // Also easy.
//       if test then foo; end;     // Alright; we can test for then.
//       if (test) foo; end;        // Undocumented, allowed syntax.
//       if test foo; end;          // Need a full parser to parse this!
//       It's a similar scenario with while and select.  The first
//       three cases are supported.  [,;] are easy to split on, and
//       "then" is easy enough to find.  The last case looks like
//       it's "if it causes a syntax error, insert a line break,"
//       and I don't want to fully duplicate Scilab's parser.
//       Actually, the fourth case requires a full parser as well,
//       to know where to split in cases such as:
//       if (a) | (b) f; else g; end;    // These are the same, valid, lines.
//       if a | b f; else g; end;        // a,b are variables.  f,g are functions.
//       
//   . Variable names are truncated after their 24th characters.
//     Not a big deal, just something to watch out for when generating
//     function names on the fly.

// Gotchas in Scilab semantics:
//
//   . Just a note, that there is no short-circuit evaluation with
//     logical operators.
//   . Nonexes, for lack of a better word.
//     A non-executed statement ("nonex") is a Scilab statement inside
//     of a function that the Scilab profiler count as being executed when
//     the function's profile() is retrieved.  We need to ignore these in
//     the counts if code is to actually achieve 100% coverage.  The
//     existing nonexes are listed in isNonex().

// Globals:
//   cover__fns(index,:) = [function name, file name]
//   cover__fnLines(index,:) = [orig start line, formatted start line,
//                              statement count, count without nonexes]
// Functions:
//   _cO_FNNAME(formattedLineNo) = orignalLineNo (short for cover__fnOrigLines_FNNAME)
//   _cN_NAME(i) = formattedLineNoOfIthNonexStmt (short for cover__fnNonexes_FNNAME)

function suffix = cover_profiledFileSuffix()
  // Returns the suffix that profiled files have appended to their filenames.
  suffix = "_coverage";
endfunction

function profiledName = cover_profiledName(filename)
  // Appends the profiled suffix to filename.
  //
  // Returns the result of cover_profileFileSuffix() appended to filename.

  profiledName = filename + cover_profiledFileSuffix();
endfunction

function cover_reset()
  // Resets internal coverage testing variables.
  //
  // This function resets the coverage tester's memory of
  // which profiled files and functions are loaded.  It
  // does not reload the original, unprofiled files.

  global cover__loadedFiles cover__fns cover__fnLines;
  cover__loadedFiles = [];
  cover__fns = [];
  cover__fnLines = [];
endfunction

function cover_newFile()
  // Register that a new profiled file has just been loaded.
  //
  // This function should be called immediately after a profiled
  // function has been loaded.  The proper sequence for loading
  // a profiled file (after calling cover_profileFile) is:
  //
  // clear cover_loadedFile;          // Optional, prevents warning messages.
  // getf("file.sci_profiled", "p");  // Load the file with profiling enabled.
  // cover_newFile();                 // Make coverage.sci register the file in memory.

  global cover__fns cover__fnLines;
  [newFns, newFnLines] = cover_loadedFile();
  if (size(cover__fns, 1) ~= -1) then
    for i = 1:size(newFns, 1)
      dups = find(cover__fns(:,1) == newFns(i,1));
      cover__fns(dups,:) = [];
      cover__fnLines(dups,:) = [];
      cover__fns = [cover__fns; newFns(i,:)];
      cover__fnLines = [cover__fnLines; newFnLines(i,:)];
    end;
  end;
endfunction

function profiledFiles = cover_profiledFiles()
  // Finds profiled files in the current directory.
  //
  // Returns a vector of absolute paths to the unprofiled
  // names of profiled files in the current directory.
  //
  // Example
  // -->listfiles();
  // !a.sci  a.sci_profiled  b.sci  c.sci_profiled!
  // -->cover_profiledFiles();
  // !a.sci  !
  // !c.sci  !

  x = listfiles("*" + cover_profiledFileSuffix());
  len = length(cover_profiledFileSuffix());
  for i = 1:size(x, 1)
    x(i) = part(x(i), 1:(length(x(i)) - len));
    x(i) = get_absolute_file_path(x(i)) + x(i);
  end;
  profiledFiles = x;
endfunction

function loadedFiles = cover_loadedFiles()
  // Returns a vector of loaded profiled files.
  //
  // Returns a vector of profiled files that have been loaded
  // and then had cover_newFile() called.  See cover_newFile()
  // for how to load a profiled function.
  //
  // THIS VECTOR SHOULD NOT BE MODIFIED.

  global cover__loadedFiles;
  loadedFiles = cover__loadedFiles;
endfunction

function b = cover_isFileProfiled(filename)
  // Checks if a file is profiled.
  //
  // cover_isFileProfiled("foo.sci") returns true if
  // foo.sci_covered exists, and false otherwise.

  [fp, err] = mopen(filename + cover_profiledFileSuffix(), "r");
  if (err == 0) then
    b = %t;
    mclose(fp);
  else
    b = %f;
  end;
endfunction

function fns = cover_loadedFns()
  // Returns a matrix containing the names of the functions that are loaded for coverage testing.
  //
  // A row [functionName, fileName] of the matrix represents
  // the function whose name is functionName and was loaded
  // from the file fileName.
  //
  // THIS MATRIX SHOULD NOT BE MODIFIED.

  global cover__fns;
  fns = cover__fns;
endfunction

function coverage = cover_results()
  // Returns a struct with coverage statistics.
  //
  // The struct has the following slots:
  //   stmtCoverage (vector):
  //       [totalStmtsExecuted, totalStmts, overallStmtCoverage]
  //   fnStmtCoverage (matrix):
  //       A row [stmtsExecuted, stmtsTotal] of the matrix represents the
  //       function in the matching row of the matrix returned from
  //       cover_loadedFns().  This matrix may be modified.

  global cover__fns;

  fnsRan = 0;
  totalFns = 0;
  fnStmtCov = [];
  for i = 1:size(cover__fns, 1)
    x = cover_fnStmtCoverage(cover__fns(i,1));
    fnStmtCov(i,:) = x;
    if (x(1) ~= 0) then fnsRan = fnsRan + 1; end;
    totalFns = totalFns + 1;
  end;

  fnCov = [fnsRan, totalFns];
  if (totalFns == 0) then
    fnCov(3) = 1;
  else
    fnCov(3) = fnsRan / totalFns;
  end;

  totalStmtCov = sum(fnStmtCov, 'r');
  if (totalStmtCov(2) == 0)
    totalStmtCov(3) = 1;
  else
    totalStmtCov(3) = totalStmtCov(1) / totalStmtCov(2);
  end;

  coverage = struct('fnStmtCoverage', fnStmtCov, ...
                    'fnCoverage', fnCov, ...
                    'stmtCoverage', totalStmtCov);
endfunction

function cover_printStmtCoverage(cov, unit, dbg)
  // Prints a statement coverage report.
  //
  // Takes a coverage struct cov returned from cover_results
  // (or fetches a fresh one is necessary) and prints a
  // summary to stdout, or to the logical unit if provided.

  global cover__fns cover__fnLines;

  if (argn(2) < 1 | ~isstruct(cov)) then
    cov = cover_results();
  end;

  if (argn(2) < 2 | unit == %f) then
    ios = getio();
    unit = ios(4);
  end;

  fnStmtCov = cov.fnStmtCoverage;
  zeroes = find(fnStmtCov(:,1) == 0 & fnStmtCov(:,1) ~= fnStmtCov(:,2));
  partials = find(fnStmtCov(:,1) ~= 0 & fnStmtCov(:,1) ~= fnStmtCov(:,2));
  completes = find(fnStmtCov(:,1) == fnStmtCov(:,2));
  [s, order1] = gsort(fnStmtCov(zeroes,2), "r", "d");
  order1 = zeroes(order1);
  [s, order2] = gsort(fnStmtCov(partials,2) - fnStmtCov(partials,1), "r", "d");
  order2 = partials(order2);
  [s, order3] = gsort(fnStmtCov(completes,2), "r", "d");
  order3 = completes(order3);
  order = [order1, order2, order3];

  fprintf(unit, "Statement Coverage Results:\n");
  fprintf(unit, "Coverage            Line  Function                  File\n");
  for i = order
    fprintf(unit, "%3d/%3d %3d (%3d%%)  %4d  %-24s  %s\n", ...
            fnStmtCov(i,1), fnStmtCov(i,2), ...
            fnStmtCov(i,1) - fnStmtCov(i,2), fnStmtCov(i,3)*100, ...
            cover__fnLines(i,1), cover__fns(i,1), cover__fns(i,2));
  end;
endfunction

function cover_printFn(name)
  // Prints a coverage report for a function.
  //
  // This includes the percentage statement coverage, the
  // number of lines missing, and the ranges of lines that
  // are missing.

  cov = cover_fnStmtCoverage(name);
  linesMissed = cover_fnLinesMissed(name);

  printf("Function %s:\n", name);
  if (cov(1) == cov(2)) then
    printf("  Complete statement coverage.\n");
  else
    printf("  %d%% statement coverage, %d/%d missing, lines:\n", cov(3)*100, cov(2)-cov(1), cov(2));
    n = length(linesMissed);
    i = 1;
    cont = "    ";
    while (i <= n)
      j = i;
      b = %t;
      while (b & j < n)
        if (linesMissed(j+1) - linesMissed(j) == 1) then
          j = j + 1;
        else
          b = %f;
        end;
      end;

      if (i == j) then
        printf("%s%d", cont, linesMissed(i));
      else
        printf("%s%d-%d", cont, linesMissed(i), linesMissed(j));
      end;
      cont = ", ";

      i = j + 1;
    end;

    if (cont ~= "    ") then
      printf("\n");
    end;
  end;
endfunction

function fnCoverage = cover_fnStmtCoverage(name)
  // Returns a row vector that provides coverage information about a function.
  //
  // The return value is [stmtsRan, totalStmts, percent].

  global cover__fns cover__fnLines;

  // Each row in a the profile matrix is [runs, cumulativeTime, complexity].
  prof = evstr(sprintf("profile(%s)", name));
  if (prof(1,1) == 0) then
    // Function wasn't called.
    stmtsRan = 0;
  else
    stmtsRan = sum(prof(:,1) ~= 0) - 1;  // Minus one for the "function ..." line.
  end;

  // This needs to be total /executable/ statements, excluding nonexes.
  i = find(cover__fns(:,1) == name);
  totalStmts = cover__fnLines(i,3) - cover__fnLines(i,4);

  if (totalStmts == 0 | stmtsRan == totalStmts) then  // Explicitly set to 1 if 100% coverage,
    percent = 1;                                      // don't risk any rounding errors, so
  else                                                // "fnCoverage(3) == 1" is an accurate test.
    percent = stmtsRan / totalStmts;
  end;
  fnCoverage = [stmtsRan, totalStmts, percent];
endfunction

function missedOrigLines = cover_fnLinesMissed(name)
  // Returns a vector containing the numbers of unexecuted
  // lines in the function with the given name.  These
  // numbers correspond to lines in the original file,
  // starting at 1.

  prof = evstr(sprintf("profile(%s)", name));
  missedStmts = find(prof(2:$,1)' == 0);
  fnOrigLines = evstr(sprintf("_cO_%s()", name));
  nonexes = evstr(sprintf("_cN_%s()", name));

  for i = 1:length(nonexes)
    missedStmts(find(missedStmts == nonexes(i))) = [];
  end;

  missedOrigLines = unique(fnOrigLines(missedStmts));
endfunction

function [filesProfiled, success, reason] = cover_profileFiles(filenames, forceProfiling)
  // Create up-to-date, profiled versions the given filenames.
  // Continues profiling the rest of the files if there is an
  // error profiling one of the files.
  //
  // filenames (row string vector): The files to profile.
  //
  // forceProfiling = false (optional boolean): If forceProfiling
  // is false or not provided, then only files whose profiled
  // versions are older then their originals are reprofiled.
  // If true, all files are reprofiled.
  //
  // filesProfiled (vector): A vector of files that were
  // actually profiled.
  //
  // success (boolean): Were all the files profiled successfully?
  //
  // reason (column string vector): The reasons why some
  // profiling failed.  To be passed to cover_internalError().

  global cover__loadedFiles;
  filesProfiled = [];
  success = %t;
  reason = [];

  if (argn(2) < 1) then
    filenames = listfiles("*.sci");
  elseif (size(filenames, 1) ~= 1) then
    filenames = matrix(filenames, 1, prod(size(filenames)));
  end;
  if (argn(2) < 2) then forceProfiling = %f; end;
  if (size(filenames,1) > size(filenames,2)) then filenames = filenames'; end;

  for filename = filenames
    baseName = filename;
    profName = filename + cover_profiledFileSuffix();

    if (~forceProfiling) then
      // Do we need to reprofile?  Not if we're profiled
      // and baseName is older than profName.
      if (cover_isFileProfiled(baseName)) then
        baseInfo = fileinfo(baseName);
        profInfo = fileinfo(profName);
        if (baseInfo(6) < profInfo(6)) then
          continue;
        end;
      end;
    end;

    [fp, err] = mopen(baseName, "r");
    if (err ~= 0) then
      success = %f;
      reason2 = sprintf("cover_coverFiles: Couldn''t open file %s for input.", baseName);
      cover_internalError(reason2);
      reason = [reason; reason2];
      continue;
    end;

    [reformatted, origLines, s, r] = cover_reformatFile(fp);
    success = success & s;
    mclose(fp);
    if (~success) then
      reason = [reason; sprintf("While profiling %s:", baseName); "    " + r];
      continue;
    end;

    [fp, err] = mopen(profName, "w");
    if (err ~= 0) then
      success = %f;
      reason2 = sprintf("cover_coverFiles: Couldn''t open new %s for writing.", profName);
      cover_internalError(reason2);
      reason = [reason; reason2];
      continue;
    end;

    [s, r] = cover_writeProfiledFile(reformatted, origLines, baseName, fp);
    success = success & s;
    mclose(fp);
    if (~success) then
      reason = [reason; sprintf("While profiling %s:\n    ", baseName); "    " + r];
      continue;
    end;

    filesProfiled($+1) = baseName;
    cover__loadedFiles = [cover__loadedFiles; get_absolute_file_path(filename) + filename];
    printf("Profiled %s.\n", baseName);
  end;
endfunction

function [reformatted, origLines, success, reason] = cover_reformatFile(fsrc)
  // Reformats an open file into lines with one statement each,
  // while keeping track of original line numbers.
  //
  // fsrc (file descriptor): The open file as returned from mopen.
  //
  // reformatted (string vector): The lines of the reformatted file,
  // with one statement each.
  //
  // origLines (vector): The original line numbers of the statements
  // in reformatted.  reformatted(i) originated from the line number
  // origLines(i) in the original file.
  //
  // success (boolean): Was the file reformatted successfully?
  //
  // reason (column string vector): Error messages, to be passed to
  // cover_internalError().

  // Tagged, repeated bits of code used in this function:
  // (Once getf is deprecated and exec can load profiled functions,
  // these tags could be replaced by inline functions.)
  // $AddChar$
  //    begOfLine = %f;
  //    line = line + ch;
  //    lineLen = lineLen + 1;
  // $AddLine$
  //    if (lineLen ~= 0) then
  //      reformatted(formattedLineCnt) = line;
  //      origLines(formattedLineCnt) = fileLineStart;
  //      formattedLineCnt = formattedLineCnt + 1;
  //    end;
  // $NewLine$
  //    line = "";
  //    lineLen = 0;
  //    begOfLine = %t;

  reformatted = [];
  origLines = [];
  success = %t;
  reason = [];

  line = "";
  lastCh = "";
  lastCode = %inf;
  lineLen = 0;
  useCh = %f;
  parChain = [];
  parCnt = 0;
  inString = %f;
  inComment = %f;
  begOfLine = %t;
  begOfLineBeforeComment = %t;
  formattedLineCnt = 1;
  fileLineCnt = 1;
  fileLineStart = 1;

  while %t
    if (useCh) then
      useCh = %f;
    else
      [n, ch] = mfscanf(fsrc, "%c");
      if (n == -1) then break; end;
    end;

    code = str2code(ch);
    nl = (code == 110 | code == 113);         // LF, CR
    white = (nl | code == 40 | code == -40);  // SPC, TAB

    if (inComment) then
      if (nl) then
        if (~begOfLineBeforeComment) then
          // $AddLine$
          if (lineLen ~= 0) then
            reformatted(formattedLineCnt) = line;
            origLines(formattedLineCnt) = fileLineStart;
            formattedLineCnt = formattedLineCnt + 1;
          end;
        end;
        // $NewLine$
        line = "";
        lineLen = 0;
        begOfLine = %t;

        inComment = %f;
        fileLineCnt = fileLineCnt + 1;
        fileLineStart = fileLineCnt;
      end;

    elseif (ch == """" | ch == "''") then
      // If we see a ' not in a string, it is a transpose if
      // it can syntactically be one.
      // Can be: )]}' ""
      // Can't be: [a-zA-Z0-9]+-/* ...
      if (ch ~= "''" | inString | ~cover_isTranspose(lastCh)) then
        // Not a transpose.
        inString = ~inString;
      end;
      // $AddChar$
      begOfLine = %f;
      line = line + ch;
      lineLen = lineLen + 1;

    elseif (inString) then;
      if (nl) then
        success = %f;
        reason = [reason; sprintf("cover_reformatFile: Parse error: Newline in middle of string.")];
        reason = [reason; sprintf(" line == ""%s""", line)];
        cover_internalError(reason);
        return;
      end;
      // $AddChar$
      begOfLine = %f;
      line = line + ch;
      lineLen = lineLen + 1;

    elseif (ch == "(") then
      parCnt = parCnt + 1;
      parChain(parCnt) = ")";
      // $AddChar$
      begOfLine = %f;
      line = line + ch;
      lineLen = lineLen + 1;

    elseif (ch == "[") then
      parCnt = parCnt + 1;
      parChain(parCnt) = "]";
      // $AddChar$
      begOfLine = %f;
      line = line + ch;
      lineLen = lineLen + 1;

    elseif (ch == "{") then
      parCnt = parCnt + 1;
      parChain(parCnt) = "]";   // {}'s are synonymous for []'s???
      // $AddChar$
      begOfLine = %f;
      line = line + ch;
      lineLen = lineLen + 1;

    elseif (ch == ")" | ch == "]" | ch == "}") then
      if (ch == "}") then
        ch = "]";
      end;
      if (parCnt == 0) then
        success = %f;
        reason = sprintf("cover_reformatFile: Unbalanced %s, not expecting any )]}s. (Line %d)", ...
                         ch, fileLineCnt);
        cover_internalError(reason);
        return;
      end;
      if (ch ~= parChain(parCnt)) then
        success = %f;
        reason = sprintf("cover_reformatFile: Unbalanced %s, expecting %s. (Line %d)", ...
                         ch, parChain(parCnt), fileLineCnt);
        cover_internalError(reason);
        return;
      end;
      parChain(parCnt) = [];
      parCnt = parCnt - 1;
      // $AddChar$
      begOfLine = %f;
      line = line + ch;
      lineLen = lineLen + 1;

    elseif ((ch == ";" | ch == ",") & parCnt == 0) then
      //printf("stmt %4d %4d: %s\n", formattedLineCnt, fileLineStart, line+ch);

      // $AddLine$
      if (lineLen ~= 0) then
        reformatted(formattedLineCnt) = line;
        origLines(formattedLineCnt) = fileLineStart;
        formattedLineCnt = formattedLineCnt + 1;
      end;

      // $NewLine$
      line = "";
      lineLen = 0;
      begOfLine = %t;

    elseif (nl) then
      if (~(lastCode == 113 & code == 110))
        fileLineCnt = fileLineCnt + 1;
        line = stripblanks(line, %t);
        len = length(line);
        b = %f;
        if (len >= 2) then b = (part(line, (len - 1):len) == ".."); end;
        if (b) then
          i = len - 1;
          while (i > 1)
            if (part(line, i-1) == ".") then
              i = i - 1;
            else
              break;
            end;
          end;
          line = part(line, 1:i-1);
          lineLen = i-1;
        else
          if (~begOfLine) then
            //printf("line %4d %4d: %s\n", formattedLineCnt, fileLineStart, line);

            // $AddLine$
            if (lineLen ~= 0) then
              reformatted(formattedLineCnt) = line;
              origLines(formattedLineCnt) = fileLineStart;
              formattedLineCnt = formattedLineCnt + 1;
            end;

            // $NewLine$
            line = "";
            lineLen = 0;
            begOfLine = %t;
          end;
          fileLineStart = fileLineCnt;
        end;
      end;

    elseif (ch == "/") then
      if (lastCh ~= "/") then
        begOfLineBeforeComment = begOfLine;
        // $AddChar$
        begOfLine = %f;
        line = line + ch;
        lineLen = lineLen + 1;
      else
        line = part(line, 1:(length(line) - 1))
        lineLen = lineLen - 1;
        inComment = %t;
      end;

    elseif (begOfLine & white) then
      // Drop.

    elseif (lineLen >= 9) then       // if x then
      // b: "is if/then"
      b = ~cover_isNameChar(ch);
      if (b) then b = ~cover_isAnyNameChar(part(line, lineLen-4)); end;
      if (b) then b = (part(line, lineLen-3:lineLen) == "then"); end;
      if (b) then
        // $AddLine$
        if (lineLen ~= 0) then
          reformatted(formattedLineCnt) = line;
          origLines(formattedLineCnt) = fileLineStart;
          formattedLineCnt = formattedLineCnt + 1;
        end;

        // $NewLine$
        line = "";
        lineLen = 0;
        begOfLine = %t;

        useCh = %t;
        continue;
      else
        // $AddChar$
        begOfLine = %f;
        line = line + ch;
        lineLen = lineLen + 1;
      end;

    elseif (~cover_isNameChar(ch) & line == "else") then
      // $AddLine$
      if (lineLen ~= 0) then
        reformatted(formattedLineCnt) = line;
        origLines(formattedLineCnt) = fileLineStart;
        formattedLineCnt = formattedLineCnt + 1;
      end;
      // $NewLine$
      line = "";
      lineLen = 0;
      begOfLine = %t;

      useCh = %t;
      continue;

    else
      // $AddChar$
      begOfLine = %f;
      line = line + ch;
      lineLen = lineLen + 1;
    end;

    lastCh = ch;
    lastCode = code;
  end;

  // $AddLine$
  if (lineLen ~= 0) then
    reformatted(formattedLineCnt) = line;
    origLines(formattedLineCnt) = fileLineStart;
    formattedLineCnt = formattedLineCnt + 1;
  end;
endfunction

function [success, reason] = cover_writeProfiledFile(src, origLines, baseName, fp)
  // Takes the resulting statements and original line numbers
  // from cover_reformatFile and writes the new, profiled file,
  // inserting functions needed for profiling: _cO_*, _cN_*,
  // and cover_loadedFile().
  //
  // src (string vector): Statements, one per entry, as returned
  // from cover_profileFiles().
  //
  // origLines (vector): The original line numbers of each of
  // the lines in src, as returned from cover_profileFiles().
  //
  // baseName (string): The name of the original file being
  // processed.  This lets the coverage tester know which file
  // a loaded function came from.
  //
  // fp (file descriptor): The file to write into, as per mopen().
  //
  // success (boolean): Was the profiled file written successfully?
  //
  // reason (column string vector): Error messages, to be passed to
  // cover_internalError().

  success = %t;
  reason = [];
  fns = [];
  fnLines = [];

  curFnName = %f;       // The current function's name.  %f if we're not
                        // currently in a function definition.
  fnCnt = 0;            // Used as an index for origLines and nonexes.
  curFnLines = [];      // [origStartLine, formattedStartLine,
                        //  executedStmtCount, nonexecutedStmtCount]
  fnOrigLines = [];     // An array mapping a function's function-relative
  fnStmtCnt = 0;        //  formatted lines to their file-relative lines in
                        //  the original file.
  fnNonexes = [];       // An array containing the function-relative formatted
  fnNonexCnt = 0;       //  lines that are non-executable statements, and are
                        //  not counted as executed in a function's profile().

  ///// Process the file, writing the reformatted functions and
  ///// keeping track of information related to the functions.

  lineCnt = size(src, 1);
  for lineNum = 1:lineCnt+1
    if (lineNum <= lineCnt) then line = src(lineNum); end;

    isFnDecl = (part(line, 1:8) == "function" & ~cover_isNameChar(part(line, 9)));

    if (lineNum == lineCnt + 1) then
      endCurFn = %t;
    else
      endCurFn = (line == "endfunction");
    end;

    if (curFnName ~= %f) then
      if (endCurFn) then
        curFnLines(3) = fnStmtCnt;
        curFnLines(4) = fnNonexCnt;

        dups = find(fns(:,1) == curFnName);
        if (length(dups) ~= 0) then
          fnCnt = fnCnt - length(dups);
          fns(dups,:) = [];
          fnLines(dups,:) = [];
          fnOrigLines(dups,:) = [];
          fnNonexes(dups,:) = [];
        end;

        fns = [fns; curFnName, baseName];
        fnLines = [fnLines; curFnLines];

        curFnName = %f;
      elseif (lineNum == lineCnt + 1) then
        break;
      else
        fnStmtCnt = fnStmtCnt + 1;
        fnOrigLines(fnCnt, fnStmtCnt) = origLines(lineNum);

        if (isNonex(line)) then
          fnNonexCnt = fnNonexCnt + 1;
          fnNonexes(fnCnt, fnNonexCnt) = fnStmtCnt;
        end;
      end;
    end;

    if (lineNum == lineCnt + 1) then break; end;

    if (isFnDecl) then
      fnCnt = fnCnt + 1;
      curFnName = cover_fnNameFromDecl(line);
      curFnLines = [origLines(lineNum), lineNum, 0, 0];
      fnOrigLines(fnCnt, 1) = origLines(lineNum);
      fnStmtCnt = 0;
      fnNonexes(fnCnt, 1) = 0;
      fnNonexCnt = 0;
    end;

    mfprintf(fp, "%s\n", src(lineNum));
  end;


  ///// The special functions below prepend a _cO_ and _cN_ to each function's
  ///// name, so if there are any functions with names longer than 20 characters,
  ///// the new names are liable to collide.  Issue an error.
  fnsTooLong = fns(find(length(fns(:,1)) > 20),1);
  if (fnsTooLong ~= []) then
    printf("cover_writeProfiledFile: Warning: Function names in %s are too long (>20 chars):\n", baseName);
    for i = 1:size(fnsTooLong, 1)
      printf("    %24s -> %20s\n", part(fnsTooLong(i), 1:24), part(fnsTooLong(i), 1:20));
    end;
  end;


  ///// Write special functions related to coverage testing.

  mfprintf(fp, "\nfunction [newFns, newFnLines] = cover_loadedFile()\n");

  // Write cover__fns.
  mod = 10;
  mfprintf(fp, "newFns = [");
  cont = "";
  for i = 1:size(fns, 1)
    if (mod == 0) then
      mfprintf(fp, "];\nnewFns = [newFns");
      cont = "; ";
      mod = 10;
    end;
    mfprintf(fp, "%s""%s"", ""%s""", cont, fns(i,1), fns(i,2));
    cont = "; ";
    mod = mod - 1;
  end;
  mfprintf(fp, "];\n");

  // Write cover__fnLines.
  mod = 10;
  mfprintf(fp, "newFnLines = [");
  cont = "";
  for i = 1:size(fns, 1)
    if (mod == 0) then
      mfprintf(fp, "];\nnewFnLines = [newFnLines");
      cont = "; ";
      mod = 10;
    end;
    mfprintf(fp, "%s%d, %d, %d, %d", cont, fnLines(i,1), fnLines(i,2), fnLines(i,3), fnLines(i,4));
    cont = "; ";
    mod = mod - 1;
  end;
  mfprintf(fp, "];\nendfunction\n");

  // Write _cO_NAME().
  for i = 1:size(fns, 1)
    mfprintf(fp, "\nfunction z = _cO_%s()\n", fns(i,1));
    mod = 30;
    mfprintf(fp, "z = [");
    cont = "";
    for j = 1:size(fnOrigLines, 2)
      x = fnOrigLines(i,j);
      if (x == 0) then break; end;
      if (mod == 0) then
        mfprintf(fp, "];\nz = [z");
        cont = ",";
        mod = 30;
      end;
      mfprintf(fp, "%s%d", cont, x);
      cont = ",";
      mod = mod - 1;
    end;
    mfprintf(fp, "];\nendfunction\n");
  end;

  // Write _cN_NAME().
  for i = 1:size(fns, 1)
    mfprintf(fp, "\nfunction z = _cN_%s()\n", fns(i,1));
    mod = 30;
    mfprintf(fp, "z = [");
    cont = "";
    for j = 1:size(fnNonexes, 2)
      x = fnNonexes(i,j);
      if (x == 0) then break; end;
      if (mod == 0) then
        mfprintf(fp, "];\nz = [z");
        cont = ",";
        mod = 30;
      end;
      mfprintf(fp, "%s%d", cont, x);
      cont = ",";
      mod = mod - 1;
    end;
    mfprintf(fp, "];\nendfunction\n");
  end;
endfunction

function name = cover_fnNameFromDecl(line)
  // Given a line of code with a function declaration,
  // extracts and returns just the function name.
  //
  // line (string): The line of code with the function declaration.
  //
  // name (string): The name of the function being defined.

  len = length(line);
  i = max([strindex(line, ["=", "]"]), min(strindex(line, " "))])

  i = i + 1;
  while (~cover_isBegOfNameChar(part(line, i)) & i < len)
    i = i + 1;
  end;
  nameStart = i;

  while %t
    if (i == len) then break; end;
    if (~cover_isNameChar(part(line, i + 1))) then break; end;
    i = i + 1;
  end;
  nameEnd = i;

  name = part(line, nameStart:nameEnd);
endfunction

function old = cover_coverErrors(interceptErrors)
  // When called with no arguments, returns whether or not
  // the coverage tester is intercepting errors.  When called
  // with a single argument, sets whether or not the coverage
  // tester is intercepting errors, possibly clearing an error
  // condition, and returns the old state.
  //
  // Calling cover_coverErrors(%t) is the preferred way to
  // clear an existing error flag.
  //
  // interceptErrors (optional boolean): If true, sets cover_error
  // to only set a flag, and initialized the error flag with [].
  // If false, sets cover_error to raise an error like usual.
  //
  // old (boolean): The old intercept state before the function was called.

  global __cover_errors __cover_errored;
  if (argn(1) >= 1) then
    old = __cover_errors;
  end;
  if (argn(2) >= 1) then
    __cover_errors = ~~interceptErrors;
    __cover_errored = [];
  end;
endfunction

function result = cover_errored()
  // Returns whether the current error flag, set in either
  // cover_coverErrors() or cover_error().

  global __cover_errored;
  result = __cover_errored;
endfunction

function cover_error(x, y)
  // If the coverage tester has been set to intercept errors
  // (see cover_coverErrors), this function will merely set
  // the error flag to [x,y].  The flag can be tested with
  // cover_errored().  If the coverage tester is not intercepting
  // errors, cover_error's arguments are passed directly to
  // error().
  //
  // Note that the default error flag is [] (which is logically
  // false), and for the purposes of being able to write
  // "if cover_errored() then", error flags should be set
  // to a logically true value with cover_error and cleared
  // with cover_coverErrors(%t).
  //
  // x = %t (optional): The first component of the error flag.
  //
  // y = [] (optional): The second component of the error flag.

  global __cover_errors __cover_errored;
  if (__cover_errors) then
    if (argn(2) < 1) then x = %t; end;
    if (argn(2) < 2) then y = []; end;
    __cover_errored = [x, y];
  else
    select (argn(2))
    case 0 then
      error("cover_error: Generic error.");
    case 1 then
      error(x);
    case 2 then
      error(x, y);
    end;
  end;
endfunction

function b = cover_isBegOfNameChar(ch)
  // Returns whether the character ch may begin a variable name.  See "help names".
  b = (length(strindex("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ%_#!$?", ch)) ~= 0)
endfunction

function b = cover_isNameChar(ch)
  // Returns whether the character ch may be in a variable name not at its beginning.  See "help names".
  b = (length(strindex("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890_#!$?", ch)) ~= 0)
endfunction

function b = cover_isAnyNameChar(ch)
  // Returns whether the character ch may appear somewhere in a variable name.  See "help names".
  b = (length(strindex("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890%_#!$?", ch)) ~= 0)
endfunction

function b = cover_isTranspose(prevCh)
  // Returns whether the character prevCh may preceed a transpose character.
  b = (length(strindex(")]abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890_#!$?", prevCh)) ~= 0)
endfunction

function b = isNonex(stmt)
  // Returns whether the statement will not be marked as executed by the Scilab profiler.
  b = (line == "return" | line == "break")
endfunction

function cover_internalError(reason)
  // Prints internal coverage testing errors to the screen.
  //
  // reason (column string vector): The strings to print,
  // one at a time, with a line break after each.

  l = prod(size(reason));
  for i = 1:l
    printf("%s\n", msg(i));
  end;
endfunction
